from __future__ import annotations

from functools import reduce
from typing import TypeVar, Generic, Tuple, Any, Dict, Set, Callable, List, Union, Iterable

from council.abstract_council import Council
from council.return_value import MemberAction

R = TypeVar('R')
R2 = TypeVar('R2')

_no_initial = object()


class ListCouncil(Council[List[R]], Generic[R]):
    def initial_result(self):
        return []

    def default_action(self, out):
        return self.Append(out)

    def reduce(self, func: Callable[[R2, R], R2], initial: R2 = _no_initial) -> Callable[..., R2]:
        """
        :return: a callaback that converts the council's output according to func (using reduction)
        """
        if initial is _no_initial:
            return self.map(
                lambda a: reduce(func, a)
            )
        else:
            return self.map(
                lambda a: reduce(func, a, initial)
            )

    class Extend(MemberAction):
        """
        Extend the result list by iterables
        """

        def __init__(self, *args: Iterable):
            self.args = args

        def __call__(self, current, state):
            state.partial_result.extend(chain.from_iterable(self.args))
            return True

    class Append(MemberAction):
        """
        Append members to the result list
        """

        def __init__(self, *args):
            self.args = args

        def __call__(self, current, state):
            state.partial_result.extend(self.args)
            return True

    class Insert(MemberAction):
        """
        Insert members into the result at index
        """

        def __init__(self, index: int, *items):
            self.index = index
            self.items = items

        def __call__(self, current, state):
            state.partial_result[self.index:self.index] = self.items
            return True

    class RemoveResult(MemberAction):
        """
        Remove elements from the result list
        """

        def __init__(self, *args):
            self.args = args

        def __call__(self, current, state):
            for a in self.args:
                try:
                    state.partial_result.remove(a)
                except ValueError:
                    raise ValueError(f'value {a} not found in results')
            return True

    class PopResult(MemberAction):
        """
        Remove elements from the result list by index
        """

        def __init__(self, *indices: Union[int, slice]):
            self.args = indices

        def __call__(self, current, state):
            for a in sorted(self.args,
                            key=lambda k: (k.stop if isinstance(k, slice) else k) % len(state.partial_result),
                            reverse=True):
                state.partial_result.pop(a)
            return True

    class ResetClass(MemberAction):
        def __call__(self, current, state) -> bool:
            state.partial_result.clear()
            return True

    Reset = ResetClass()



from __future__ import annotations

from abc import abstractmethod
from functools import partial, update_wrapper, lru_cache
from typing import Set, Iterable, TypeVar, Callable, Generic, Tuple, Dict, Any

from council.return_value import MemberAction
from council.council_member import CouncilMember

R = TypeVar('R')
R2 = TypeVar('R2')


class AbstractCouncil(Generic[R]):
    @abstractmethod
    def __call__(self, *args, **kwargs) -> R:
        pass

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return partial(self, instance)

    @abstractmethod
    def add_member(self, member):
        pass

    @abstractmethod
    def remove_member(self, member):
        pass

    def on_member_change(self, member):
        pass

    def join_temporary(self, member):
        """
        add a member as add_member to the council, and get a callback that removes it
        :return: a callable that removes the new member from the council. Has a value __member__ that stores the member.
        """
        member = self.add_member(member)
        ret = partial(self.remove_member, member)
        ret.__member__ = member
        return ret

    def map(self, func: Callable[[R], R2]) -> MappedCouncil[R, R2]:
        """
        :return: a callaback that converts the council's output according to func
        """
        return MappedCouncil(self, func)

    def lru_cache(self, *args, **kwargs):
        return CachedCouncil(self, *args, **kwargs)


class CouncilCallState(Generic[R]):
    """
    Representing a state of calling a council object, passed to members as they are called
    """

    def __init__(self, council: Council[R], args: Tuple, kwargs: Dict[str, Any]):
        self.council = council
        self.args = args
        self.kwargs = kwargs

        self.dependency_stack = []
        self.pending_members = set(self.council.members)
        self.partial_result = self.council.initial_result()

    def call_next(self):
        """
        process a member of the council

        :return: whether to continue with the call
        """
        if self.dependency_stack:
            member = self.dependency_stack.pop()
        elif self.pending_members:
            member = self.pending_members.pop()
        else:
            return False

        out = member.call(self.args, self.kwargs, self)
        if not isinstance(out, MemberAction):
            out = self.council.default_action(out)
        return out(member, self)

    def __call__(self) -> R:
        while self.call_next():
            pass
        return self.partial_result


class Council(AbstractCouncil[R], Generic[R]):
    def __init__(self, name: str = None, decorators=()):
        """
        :param name: the name of the council
        :param decorators: these decorators will be applied, in reverse order, to all members added
        """
        self.members: Set[CouncilMember[R]] = set()
        if not isinstance(decorators, Iterable):
            decorators = decorators,
        self.member_wrappers = decorators
        if name:
            self.__name__ = name

    @classmethod
    def from_template(cls, template=None, **kwargs):
        """
        will create a council based on a wrapped callable, usable as decorator
        :param template: the object to base on, most likely a function. Will never be called.
        :param kwargs: forwarded to Council.__init__
        """
        if template is None:
            return partial(cls.from_template, **kwargs)

        name = getattr(template, '__name__', None)
        ret = cls(name, **kwargs)
        update_wrapper(ret, template)
        return ret

    def __set_name__(self, owner, name):
        ex_name = getattr(self, '__name__', None)
        if ex_name is None:
            self.__name__ = name
        elif ex_name != name:
            raise ValueError(f'this council already has an assigned name {ex_name}')

    def call_state(self, args, kwargs):
        """
        Generate a new call state for the council
        """
        return CouncilCallState(self, args, kwargs)

    @abstractmethod
    def initial_result(self) -> R:
        pass

    @abstractmethod
    def default_action(self, out) -> MemberAction:
        pass

    def __call__(self, *args, **kwargs) -> R:
        call_state = self.call_state(args, kwargs)
        return call_state()

    def add_member(self, member):
        """
        Add a member, applying all the decorators, can be used as decorator

        :param member: the object to convert to member, decorate, and add to the council
        :return: the member added, including all conversions and decorations
        """
        if not isinstance(member, CouncilMember):
            member = CouncilMember.coerce(member)
        for ad in reversed(self.member_wrappers):
            member = ad(member)
        self.members.add(member)
        member.introduce(self)
        return member

    def remove_member(self, member):
        """
        remove a member from the council
        """
        self.members.remove(member)
        member.depart(self)

    def __repr__(self):
        try:
            return f'{type(self).__name__}({self.__name__!r})'
        except AttributeError:
            return super().__repr__()

    class Modify(MemberAction):
        def __init__(self, func):
            self.func = func

        def __call__(self, current, state) -> bool:
            state.partial_result = self.func(state.partial_result)
            return True

    class Mutate(MemberAction):
        def __init__(self, func):
            self.func = func

        def __call__(self, current, state) -> bool:
            self.func(state.partial_result)
            return True

    class ResetClass(MemberAction):
        def __call__(self, current, state) -> bool:
            state.partial_result = state.council.initial_result()
            return True

    Reset = ResetClass()


class MappedCouncil(AbstractCouncil[R2], Generic[R, R2]):
    def __init__(self, council: AbstractCouncil[R], conv: Callable[..., R2]):
        self.council = council
        self.conv = conv

    def __call__(self, *args, **kwargs):
        ret = self.council(*args, **kwargs)
        return self.conv(ret)

    def __repr__(self):
        return f'{self.council!r}.map({self.conv!r})'

    def add_member(self, member):
        return self.council.add_member(member)

    def remove_member(self, member):
        return self.council.remove_member(member)

    def on_member_change(self, member):
        return self.council.on_member_change(member)


class CachedCouncil(Generic[R], AbstractCouncil[R]):
    """
    A council that caches its results. Note that the council returns tuples, not lists.
    """

    def __init__(self, inner: AbstractCouncil[R], *args, **kwargs):
        self.inner = inner
        self.cache = lru_cache(*args, **kwargs)(inner)

    def __call__(self, *args, **kwargs) -> Tuple[R]:
        return self.cache(*args, **kwargs)

    def add_member(self, member):
        ret = super().add_member(member)
        self.cache_clear()
        return ret

    def remove_member(self, member):
        ret = super().remove_member(member)
        self.cache_clear()
        return ret

    def cache_clear(self):
        self.cache.cache_clear()

    def on_member_change(self, member):
        self.cache_clear()
        return self.inner.on_member_change(member)



from __future__ import annotations

from abc import abstractmethod
from itertools import chain
from typing import Iterable

from council.council_member import CouncilMember


class MemberAction:
    """
    A member action is an object that, when returned by a member, may influence the computation of the council
    """

    @abstractmethod
    def __call__(self, current, state) -> bool:
        """
        is called when returned by a member

        :param current: the member that returned self
        :param state: the CallState
        :return: whether to continue the computation
        """
        pass

    def __add__(self, other):
        if not isinstance(other, MemberAction):
            other = DefaultAction(other)
        return Joined((self, other))

    def __radd__(self, other):
        other = DefaultAction(other)
        return other + self


class ContinueClass(MemberAction):
    """
    Ignore the member and add nothing to the result
    """

    def __call__(self, *args, **kwargs):
        return True


Continue = ContinueClass()


class BreakClass(MemberAction):
    """
    End the computation, no other members will be processed
    """

    def __call__(self, current, state):
        return False


Break = BreakClass()


class Postpone(MemberAction):
    """
    Request that the member be called again after certain other members have been called
    """

    def __init__(self, *wait_for: CouncilMember):
        if not wait_for:
            raise ValueError('must specify a member to wait for')
        self.wait_for = wait_for

    def __call__(self, current, state):
        if current not in state.dependency_stack:
            state.dependency_stack.append(current)
        for wf in self.wait_for:
            try:
                state.pending_members.remove(wf)
            except KeyError:
                raise KeyError(f'cannot wait for member outside of the pending set {wf}')
            state.dependency_stack.append(wf)
        return True


class Enqueue(MemberAction):
    """
    Add other members to the council call. Note that the members are not added to the council
    """

    def __init__(self, *args: CouncilMember):
        self.args = args

    def __call__(self, current, state):
        state.pending_members.update(self.args)
        return True


class DefaultAction(MemberAction):
    def __init__(self, val):
        self.val = val

    def __call__(self, current, state):
        return state.council.default_action(self.val)(current, state)


class Joined(MemberAction):
    """
    Perform multiple member actions
    """

    def __init__(self, parts: Iterable[MemberAction]):
        self.parts = parts

    def __call__(self, *args, **kwargs):
        ret = True
        for p in self.parts:
            ret &= p(*args, **kwargs)
        return ret

    def __add__(self, other):
        if isinstance(other, type(self)):
            return type(self)(chain(self.parts, other.parts))
        if not isinstance(other, MemberAction):
            other = DefaultAction(other)
        return type(self)((*self.parts, other))


__all__ = ['Continue', 'Break', 'Postpone', 'Enqueue', 'MemberAction']



from __future__ import annotations

from typing import TypeVar, Generic, Callable, Union, TYPE_CHECKING
from abc import abstractmethod
from functools import update_wrapper
from inspect import signature

if TYPE_CHECKING:
    from council import MemberAction

R = TypeVar('R')


class CouncilMember(Generic[R]):
    """
    A council member that can participate in councils
    """

    def __init__(self):
        self.councils = set()

    @abstractmethod
    def call(self, args, kwargs, council_state) -> Union[R, MemberAction]:
        """
        called when the member is processed

        :param args: arguments passed to the council
        :param kwargs: keyword arguments passed to the council
        :param council_state: the CallState instance of the current council call
        :return: a value to append to the council's result or a MemberAction to issue special instructions
        """
        pass

    @classmethod
    def coerce(cls, x):
        """
        convert x to a CouncilMember, if needed
        """
        if isinstance(x, cls):
            return x
        if isinstance(x, Callable):
            return council_member(x)
        else:
            raise TypeError(f'cannot coerce {x} to a council member')

    def send_modify(self):
        for c in self.councils:
            c.on_member_change(self)

    def introduce(self, council):
        if council in self.councils:
            raise Exception('this member is already in the council')
        self.councils.add(council)

    def depart(self, council):
        self.councils.remove(council)


class FuncMember(CouncilMember[R], Generic[R]):
    """
    A council member that wraps a function
    """

    def __init__(self, func):
        super().__init__()
        self.func = func
        update_wrapper(self, self.func)

    def call(self, args, kwargs, council_state):
        return self(*args, council_state=council_state, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return f'{type(self).__name__}({self.func!r})'


class SimpleFuncMember(FuncMember[R], Generic[R]):
    """
    A FuncMember that ignores the council state provided
    """

    def call(self, args, kwargs, council_state=None):
        return self(*args, **kwargs)


def council_member(func: Callable[..., R]) -> CouncilMember[R]:
    """
    a function decorator to create an appropriate member

    :param func: the function to wrap
    :return: a CouncilMember that will call func, ignoring council_state unless a parameter of that name exists
     in func's signature
    """
    sig = signature(func)
    if 'council_state' in sig.parameters:
        return FuncMember(func)
    return SimpleFuncMember(func)



from __future__ import annotations

from enum import Enum, auto
from typing import TypeVar, Generic, Dict, Iterable, Sized

from council.abstract_council import Council
from council.return_value import MemberAction

K = TypeVar('K')
V = TypeVar('V')

_no_default = object()


class OnExistAction(Enum):
    Update = auto()
    Skip = auto()
    Raise = auto()


class DictCouncil(Council[Dict[K, V]], Generic[K, V]):
    def initial_result(self):
        return {}

    def default_action(self, out):
        if not isinstance(out, Sized) or not isinstance(out, Iterable) or len(out) != 2:
            raise TypeError('raw return values must be a 2-iterable')
        return self.Set(*out, on_exist=OnExistAction.Raise)

    class Set(MemberAction):
        def __init__(self, key, value, *, on_exist=OnExistAction.Update):
            self.key = key
            self.value = value
            self.on_exist = on_exist

        def __call__(self, current, state):
            d = state.partial_result
            if self.key in d:
                if self.on_exist == OnExistAction.Raise:
                    raise Exception(f'result already contains key {self.key}')
                elif self.on_exist == OnExistAction.Skip:
                    pass
                elif self.on_exist == OnExistAction.Update:
                    d[self.key] = self.value
                else:
                    raise Exception('bad on_exist value')
            else:
                d[self.key] = self.value
            return True

    # noinspection PyPep8Naming
    @classmethod
    def SetDefault(cls, *args):
        return cls.Set(*args, on_exist=OnExistAction.Skip)

    class ModifyValue(MemberAction):
        def __init__(self, key, func, default=_no_default):
            self.key = key
            self.func = func
            self.default = default

        def __call__(self, current, state):
            d = state.partial_result
            old_val = d.get(self.key, _no_default)
            if old_val is _no_default:
                if self.default is _no_default:
                    raise KeyError(self.key)
                new_val = self.default
            else:
                new_val = self.func(old_val)
            old_val[self.key] = new_val
            return True

    class MutateValue(MemberAction):
        def __init__(self, key, func, default=_no_default):
            self.key = key
            self.func = func
            self.default = default

        def __call__(self, current, state):
            d = state.partial_result
            old_val = d.get(self.key, _no_default)
            if old_val is not _no_default:
                self.func(old_val)
            elif self.default is _no_default:
                raise KeyError(self.key)
            else:
                d[self.key] = self.default
            return True

    class Pop(MemberAction):
        def __init__(self, *keys, allow_missing=False):
            self.keys = keys
            self.allow_missing = allow_missing

        def __call__(self, current, state):
            for k in self.keys:
                try:
                    del state.partial_result[k]
                except KeyError:
                    if not self.allow_missing:
                        raise
            return True

    class Update(MemberAction):
        def __init__(self, iterable=(), **kwargs):
            self.map = iterable
            self.kwargs = kwargs

        def __call__(self, current, state):
            state.partial_result.update(self.map, **self.kwargs)
            return True

    class ResetClass(MemberAction):
        def __call__(self, current, state) -> bool:
            state.partial_result.clear()
            return True

    Reset = ResetClass()



from abc import ABC, abstractmethod

from council.council_member import CouncilMember
from council.return_value import Postpone, Continue, Break, MemberAction


class MemberWrapper(CouncilMember):
    def __init__(self, inner: CouncilMember):
        super().__init__()
        self.inner = CouncilMember.coerce(inner)

    def call(self, args, kwargs, council_state):
        return self.inner.call(args, kwargs, council_state)

    def __repr__(self):
        return f'{type(self).__name__}({self.inner!r})'


class AlwaysAfter(MemberWrapper):
    def __init__(self, inner, after):
        super().__init__(inner)
        self.after = after

    def call(self, args, kwargs, council_state):
        intersect = council_state.pending_members.intersection(self.after)
        if intersect:
            return Postpone(*intersect)
        return super().call(args, kwargs, council_state)

    def __repr__(self):
        return f'{type(self).__name__}({self.inner!r}, {self.after!r})'


class AlwaysLast(MemberWrapper):
    def call(self, args, kwargs, council_state):
        other_member = next(iter(council_state.pending_members), None)
        if other_member:
            return Postpone(other_member)
        return super().call(args, kwargs, council_state)


class AutoContinues(MemberWrapper, ABC):
    @abstractmethod
    def is_continue(self, x) -> bool:
        pass

    def call(self, args, kwargs, council_state):
        ret = super().call(args, kwargs, council_state)
        if self.is_continue(ret):
            ret = Continue
        return ret


class NoneContinues(AutoContinues):
    def is_continue(self, x) -> bool:
        return x is None


class TruthContinues(AutoContinues):
    def is_continue(self, x) -> bool:
        return x


class NonTruthContinues(AutoContinues):
    def is_continue(self, x) -> bool:
        return not x


class AutoBreaks(MemberWrapper, ABC):
    @abstractmethod
    def is_break(self, x) -> bool:
        pass

    def call(self, args, kwargs, council_state):
        ret = super().call(args, kwargs, council_state)
        if self.is_break(ret):
            ret = ret + Break
        return ret


class NotNoneBreaks(AutoBreaks):
    def is_break(self, x) -> bool:
        return x is not None


class TruthBreaks(AutoBreaks):
    def is_break(self, x) -> bool:
        return x and not isinstance(x, MemberAction)


class NonTruthBreaks(AutoBreaks):
    def is_break(self, x) -> bool:
        return not x


always_last = AlwaysLast

none_continues = NoneContinues
truth_continues = TruthContinues
non_truth_continues = NonTruthContinues

not_none_breaks = NotNoneBreaks
truth_breaks = TruthBreaks
non_truth_breaks = NonTruthBreaks


def always_after(*members):
    def ret(member):
        return AlwaysAfter(member, members)

    return ret


__all__ = [
    'always_last', 'always_after',
    'none_continues', 'truth_continues', 'non_truth_continues',
    'not_none_breaks', 'truth_breaks', 'non_truth_breaks'
]



from council.abstract_council import Council

from council.list_council import ListCouncil
from council.council_member import council_member, CouncilMember
from council.return_value import *
from council.member_wrapper import *



import setuptools

import council_data as council

setuptools.setup(
    name='council',
    version=council.__version__,
    url=council.__url__,
    author=council.__author__,
    packages=['council', 'council_data'],
    python_requires='>=3.8.0',
    include_package_data=True,
    data_files=[
        ('', ['README.md', 'CHANGELOG.md']),
    ],
    install_requires=[],
)
